import get from 'lodash.get';
import { ValidationError } from 'yup';
import { EFieldTypes, IForm, IFormField, IFormGroup, IValidationError, IValidationErrorField } from './models';
import { getFieldError } from './utils';

/**
 * @param jsonSchema - jsonSchema validation object
 * @param t - i18n translate function
 * @returns {any}
 * @description getErrorMessages function returns an object with the error messages for each field in the form based on the jsonSchema
 */
export const getErrorMessages = (jsonSchema: any, t?: Function) => {
  const errMessages: any = {};

  Object.keys(jsonSchema.properties).forEach((property) => {
    errMessages[property] = {};
    Object.keys(jsonSchema.properties[property]).forEach((key) => {
      const value = jsonSchema.properties[property][key];
      errMessages[property][key] = getFieldError(key, value, t);
      if (Object.keys(jsonSchema).includes(key)) {
        errMessages[property].required = getFieldError('required', null, t);
      }
    });
  });

  return errMessages;
};

/**
 * @param translateKey
 * @param param
 * @returns {IValidationError}
 * @description createErrorObject function returns an object with the error message
 * for a field in the form in a proper format
 * @example createErrorObject('required', { param: 'email' }) => { translationKey: 'required', fallbackLanguageTranslation: 'This field is required', meta: { param: 'email' } }
 */
export const createErrorObject = (translateKey: string, param?: { [key: string]: any; }): IValidationError => ({
  translationKey: translateKey, fallbackLanguageTranslation: translateKey, meta: param || {},
});

export function isFormGroup(
  toBeDetermined: IFormField | IFormGroup,
): toBeDetermined is IFormGroup {
  return !!(toBeDetermined as IFormGroup).children;
}

/**
 * @param stringifiedFunction
 * @returns {Function}
 * @description functionFromString returns a function from a stringified function
 * @example functionFromString('/Function((values) => {return values.medicalConditions.toString() === 'true'})/')
 * => (values) => values.medicalConditions.toString() === 'true'
 */
export const functionFromString = (stringifiedFunction: string) => {
  try {
    const substr = stringifiedFunction.substring(
      10,
      stringifiedFunction.length - 2,
    );
    return (0, eval)('(' + substr + ')');
  } catch (e) {
    console.warn('Error parsing function in functionFromString: ', e);
    return () => false;
  }
};

interface IValidationErrors {
  [key: string]: IValidationErrorField;
}

/**
 * @param validationError - yup validation error
 * @param initialErrors - initial errors object to be merged with the yup validation errors
 * @returns {IValidationErrors} - returns an object with the errors for each field in the form
 * @description formatYupValidationErrors function returns an object with the errors for each field in the form
 * based on the yup validation error formatting to be used in the GenericForm component
 * @example formatYupValidationErrors(yupValidationError, { email: { property: 'email', constraints: { required: { translationKey: 'required', fallbackLanguageTranslation: 'This field is required', meta: {} } } } })
 */
export const formatYupValidationErrors = (validationError: ValidationError, initialErrors: IValidationErrors = {}): IValidationErrors => {
  const errors: IValidationErrors = { ...initialErrors };

  for (const propertyValidationError of validationError.inner) {
    let errorType = propertyValidationError.type;

    if (!errorType && propertyValidationError.params?.regex) {
      if (!propertyValidationError.params?.regex) {
        continue;
      }

      errorType = 'pattern';
    }

    if (!get(errors, `${propertyValidationError.path}`)) {
      errors[propertyValidationError.path as string] = {
        property: propertyValidationError.path as string,
        constraints: {},
      };
    }

    errors[propertyValidationError.path as string].constraints[
      errorType as string
      ] = {
      ...getFieldError(
        errorType as string,
        get(propertyValidationError, 'path', null),
        undefined,
        get(propertyValidationError.params, `${propertyValidationError.type}`),
      ),
      fallbackLanguageTranslation: propertyValidationError.message,
    };
  }

  return errors;
};

/**
 * @param form - form json schema
 * @param type - field type
 * @param acc - accumulator, used for recursion
 * @returns {string[]} - array of field names
 * @description getFieldNamesByType function returns an array of field names based on the field type
 * @example getFieldNamesByType(form, EFieldTypes.TEXT) => ['firstName', 'lastName']
 */
export const getFieldNamesByType = (form: IForm | IFormGroup, type: EFieldTypes, acc: string[] = []): string[] => {
  return form.children.reduce((acc: string[], child) => {
    if (isFormGroup(child)) {
      return getFieldNamesByType(child, type, acc);
    }

    if (child.type === type) {
      acc.push(child.fieldName);
    }

    return acc;
  }, acc);
};

/**
 * @param obj - object to trim
 * @returns {any} - object with trimmed values
 * @description trimValuesInObject function returns an object with trimmed values
 */
export const trimValuesInObject = (obj: any) => {
  const trimmedValues: any = {};
  Object.keys(obj).forEach(k =>
    obj[k] = typeof obj[k] == 'string'
      ? trimmedValues[k] = String(obj[k]).trim()
      : trimmedValues[k] = obj[k]);
  return trimmedValues;
};
